/*
 * NATDet - NAT Detection tool
 * ---------------------------
 * Copyright (c) 2003, 2004 Marcin Ulikowski <elceef@itsec.pl>
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/types.h>
#ifdef UNIX
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#else
#pragma comment (lib, "WSOCK32.lib")
#pragma comment (lib, "wpcap.lib")
#endif
#include <pcap.h>
#include <signal.h>
#include <time.h>

#include "config.h"
#include "struct.h"

#ifdef UNIX
/* some ANSI codes ... */
#define CLR  "\033[2J"       /* clear screen */
#define RST  "\033[0m"       /* reset */
#define CRN  "\033[0H"       /* go to screen corner */
#define DWN  "\033[25;0H"    /* go down to 25th line */
#define YNB  "\033[1;33;44m" /* yellow foreground, blue background */
#endif

/* factors */
#define OSGENRE 0x01
#define TSTAMP  0x02
#define TTL     0x04

static struct nat_cache nat[NAT_CACHE]; /* NAT cache entries */
static struct fp_cache fp[MAX_SIGS];    /* OS fingerprints */
static struct config cfg;               /* configuration */

static u_char *iface, *rule;            /* device and BPF rule */
static u_char fhlen;                    /* frame header length */

static u_short cpos, sigs;

struct pcap_pkthdr header;
struct pcap_stat pstat;


static void bug(u_char *format, ...) { 
  u_char buffer[256];
  va_list args;

  va_start(args, format);
  vsnprintf(buffer, sizeof(buffer) - 1, format, args);
  fprintf(stderr, "[!] %s", buffer);
  va_end(args);

  exit(1);
}


/* find out frame header length */
static void set_fhlen(u_int dlink_type) {
  switch(dlink_type) {

    case DLT_RAW:        break;

    case DLT_SLIP:       fhlen = 16;
                         break;

    case DLT_EN10MB:     fhlen = 14;
                         break;

#ifdef DLT_LOOP
    case DLT_LOOP:
#endif
    case DLT_NULL:
    case DLT_PPP:        fhlen = 4;
                         break;

    case DLT_IEEE802:    fhlen = 22;
                         break;

#ifdef DLT_PPP_ETHER
    case DLT_PPP_ETHER:  fhlen = 8;
                         break;
#endif

#ifdef DLT_PFLOG
    case DLT_PFLOG:      fhlen = 28;
                         break;
#endif

#ifdef DLT_LINUX_SLL
    case DLT_LINUX_SLL:  fhlen = 16;
                         break;
#endif

#ifdef DLT_SLIP_BSDOS
    case DLT_SLIP_BSDOS: fhlen = 24;
                         break;
#endif

#ifdef DLT_PPP_BSDOS
    case DLT_PPP_BSDOS:  fhlen = 24;
                         break;
#endif

#ifdef DLT_PPP_SERIAL
    case DLT_PPP_SERIAL: break;
#endif

    default:             fprintf(stderr, "[i]: unknown datalink type (%u)!\n",
                                 dlink_type);
                         break;
  }

}


static u_char *osfp(u_short window, u_char ttl, u_short df, u_short plen) {
  u_int match, i;

  for (i = 0; i < MAX_SIGS; i++) {
    
    match = 0;

    if (fp[i].win == window) {

      if (fp[i].len == plen || !fp[i].len) {

        if (fp[i].ttl == 32 && ttl <= 32) {  /* (0; 32> */
          if ((df == IP_DF && fp[i].df == 1) || (!df && !fp[i].df))
            match = 1;
        }

        if (fp[i].ttl == 64 && ttl <= 64 && ttl > 32) {  /* (32; 64> */
          if ((df == IP_DF && fp[i].df == 1) || (!df && !fp[i].df))
            match = 1;
        }

        if (fp[i].ttl == 128 && ttl <= 128 && ttl > 64) {  /* (64; 128> */
          if ((df == IP_DF && fp[i].df == 1) || (!df && !fp[i].df))
            match = 1;
        }

        if (fp[i].ttl == 255 && ttl > 128) {  /* (128; 255> */
          if ((df == IP_DF && fp[i].df == 1) || (!df && !fp[i].df))
            match = 1;
        }

        if (match) return fp[i].os;

        continue;
      }
    }
  }

  return "Unknown";
}


static u_char *showtime(void) {
  static char buff[20];
  char *pbuff = buff;
  time_t i = time(0);
  struct tm *t;

  t = localtime((void *)&i);
  snprintf(buff, sizeof(buff), "%02u-%02u-%u %02u:%02u:%02u", t->tm_mday,
           t->tm_mon, t->tm_year+1900, t->tm_hour, t->tm_min, t->tm_sec);

  return pbuff;
}


static void load_signatures(void) {
  FILE *osprints;
  FILE *test_osprints;

  u_int osdf, osttl, osplen, oswin,
        count = 0;
  u_int test_osdf, test_osttl, test_osplen, test_oswin,
        test_count = 0;
	  
  u_char line[SIG_LEN], osname[32];
  u_char test_line[SIG_LEN];

  if (!(osprints = fopen(FP_DB, "r")))
    if (!(osprints = fopen(FP_DIR"/"FP_DB, "r")))
      bug("unable to open \"%s\"\n", FP_DB);

  while (count++ <= 65535) {
 
    fgets(line, SIG_LEN, osprints);

    if (feof(osprints)) break;

    if (line[0] == '#' || line[0] == '\n')
      continue;

    if (sscanf(line, "%u:%u:%u:%u:%[^:]",
               &oswin, &osttl, &osdf, &osplen, osname) != 5) {
      bug("syntax error on line %d in \"%s\"\n", count, FP_DB);
    }

    /* test signature */
    if (!(test_osprints = fopen(FP_DB, "r")))
      if (!(test_osprints = fopen(FP_DIR"/"FP_DB, "r")))
        bug("unable to open \"%s\"\n", FP_DB);

    while (test_count++ <= MAX_SIGS) {

      fgets(test_line, SIG_LEN, test_osprints);

      if (feof(test_osprints)) break;

      if (test_line[0] == '#' || test_line[0] == '\n')
        continue;

      if (sscanf(test_line, "%u:%u:%u:%u:",
          &test_oswin, &test_osttl, &test_osdf, &test_osplen) != 4)
        bug("syntax error on line %d in \"%s\"\n", test_count, FP_DB);

      if (test_oswin  == oswin  &&
          test_osttl  == osttl  &&
          test_osdf   == osdf   &&
          test_osplen == osplen &&
          test_count  != count) {
        bug("identical signatures on lines %u and %u in \"%s\"\n",
            count, test_count, FP_DB);
      }

    }

    fclose(test_osprints);
    /* signature tested */

    if (oswin > 65535)
      bug("WSS (%d) too high on line %u in \"%s\"\n", oswin, count, FP_DB);

    if (osttl % 32 && osttl != 255)
      bug("unusual TTL %d on line %u in \"%s\"\n", osttl, count, FP_DB);

    if (osdf > 1)
      bug("wrong DF (%d) on line %u in \"%s\"\n", osdf, count, FP_DB);

    if (osplen > 64)
      bug("packet length too high on line %u in \"%s\"\n", count, FP_DB);

    if (!(strlen(osname) - 1))
      bug("empty OS description on line %u in \"%s\"\n", count, FP_DB);

    if (sigs > MAX_SIGS) break;

    osname[strlen(osname) - 1] = '\0';

    fp[sigs].win = oswin;
    fp[sigs].ttl = osttl;
    fp[sigs].len = osplen;
    fp[sigs].os  = strdup(osname);
    if (osdf) fp[sigs].df = 1;
    else fp[sigs].df = 0;

    sigs++;
  }

  fclose(osprints);
}


static void watch(u_char *args, struct pcap_pkthdr *header, u_char *packet) {
  struct iphdr  *ip  = (struct iphdr *)(packet + fhlen);
  struct tcphdr *tcp = (struct tcphdr *)(packet + fhlen + (ip->ihl * 4));

  u_char *saddr, *optp;
#ifdef UNIX
  u_char *iaddr;
  u_char hbar[44], host[16];
#endif

  static u_char ttl_diff, ttl1, ttl2,
                osgenre[16];

  u_char fcount = 0,
         factor = 0;

  u_short oldprob, tmpprob, i, k,
          osint = 0;

  u_int timestamp = 0;

  u_long cur_time = 0;

#if BYTE_ORDER == LITTLE_ENDIAN
  ip->tot_len  = htons(ip->tot_len);
  ip->frag_off = htons(ip->frag_off);

  tcp->window  = htons(tcp->window);
#endif

  saddr = (u_char *)&ip->saddr;

  optp  = (u_char *)(tcp + 1);

  if (ip->ihl < 5 && ip->version != 0x04) {
    printf("[!] Damaged packet received\n");
    goto end;
  }

  for (i = 40; i < ip->tot_len; i++) {
    if (*(optp++) == TCP_TIMESTAMP) {
      memcpy(&timestamp, optp+1, 4);
#if BYTE_ORDER == LITTLE_ENDIAN
      timestamp = ntohl(timestamp);
#endif
      /* we assume timstamp should be longer than 10 minutes */
      if (timestamp > 65535) break;
    }
  }

  strncpy(osgenre, osfp(tcp->window, ip->ttl, ip->frag_off, ip->tot_len),
          sizeof(osgenre));

  /* make OSGENRE ID */
  for (k = 0; k < strlen(osgenre); k++)
    osint += osgenre[k] * k;

  cur_time = time(0);

  if (ip->ttl > 0 && ip->ttl <= 32) {
    ttl1 = 32 - ip->ttl;

  } else if (ip->ttl > 32 && ip->ttl <= 64) {
    ttl1 = 64 - ip->ttl;

  } else if (ip->ttl > 64 && ip->ttl <= 128) {
    ttl1 = 128 - ip->ttl;

  } else if (ip->ttl > 128)
    ttl1 = 255 - ip->ttl;

  if (ip->saddr == ip->daddr) goto end;

  for (i = 0; i < NAT_CACHE; i++) {

    if (nat[i].ip != ip->saddr) continue;

    if (nat[i].ttl > 0 && nat[i].ttl <= 32) {
      ttl2 = 32 - nat[i].ttl;

    } else if (nat[i].ttl > 32 && nat[i].ttl <= 64) {
      ttl2 = 64 - nat[i].ttl;

    } else if (nat[i].ttl > 64 && nat[i].ttl <= 128) {
      ttl2 = 128 - nat[i].ttl;

    } else if (nat[i].ttl > 128)
      ttl2 = 255 - nat[i].ttl;

    if (ttl2 > ttl1) {
      ttl_diff = ttl2 - ttl1;
    } else {
      ttl_diff = ttl1 - ttl2;
    }

    oldprob = nat[i].prob;
    tmpprob = 0;

    /* do not count "Unknown" (ID=2352) systems */
    if (osint != 2353) {
      for (k = 0; k < OS_LIMIT; k++) {
        if (nat[i].oslist[k] == osint)
          break;
	/* add another OS to the counter */
        if (!nat[i].oslist[k]) {
          nat[i].oslist[k] = osint;
          nat[i].oscount++;
          break;
        }
      }
    }

    /* refresh time */
    if (!strcmp(nat[i].osgenre, osgenre) && nat[i].ttl == ip->ttl)
      nat[i].time = cur_time;

    /* if cache entry is too old, rewrite it with new data */
    if (cur_time - nat[i].time >= LIVE_TIME) {
      nat[i].time    = cur_time;
      nat[i].ttl     = ip->ttl;
      nat[i].tstamp  = timestamp;
      nat[i].delay   = 0;
      nat[i].prob    = 0;
      nat[i].oscount = 1;
      strncpy(nat[i].osgenre, osgenre, sizeof(nat[i].osgenre));
      memset(nat[i].oslist, 0, 2*OS_LIMIT); 
      nat[i].oslist[0] = osint;
      goto end;
    }

    /* if osgenres differs and both are known, use this factor */
    if (strcmp(nat[i].osgenre, osgenre) && strcmp(osgenre, "Unknown") &&
        strcmp(nat[i].osgenre, "Unknown")) {
      tmpprob += 100;
      factor |= OSGENRE;
      fcount++;
    }

    /* if timestamp is lower than cached, use this factor */
    if (timestamp != nat[i].tstamp &&
       (timestamp + TSTAMP_DIFF < nat[i].tstamp ||
       (!nat[i].tstamp && timestamp) || (!timestamp && nat[i].tstamp))) {

      tmpprob += 100;

      /* don't forget to refresh tcp-timestamp if osgenre is the same */
      if (timestamp && !strcmp(nat[i].osgenre, osgenre))
        nat[i].tstamp = timestamp;

      factor |= TSTAMP;
      fcount++;
    }

    /* TTL difference? use this factor! */
    if (ttl_diff > 0 && ttl_diff <= HOP_DIST) {
      tmpprob += 100;
      factor |= TTL;
      fcount++;
    }

    if (fcount) {
      tmpprob = tmpprob / 3;
      /* TTL difference and/or 4 different OSes make us sure */
      if (factor & TTL || nat[i].oscount == OS_LIMIT) {
        nat[i].prob = 100;
      } else {
        /* calculate probability for the host */
        nat[i].prob = (tmpprob * (LIVE_TIME - (cur_time - nat[i].time))
                      / LIVE_TIME + oldprob) / 2;
      }
    }

#ifdef UNIX
    if (cfg.iactive) {
      /* For those who just hate reading logs :)) */
      printf("%s%sNATDet \"I-hate-logs!\" mode   %s   cache: %3u%%  %16s%s\n",
             CRN, YNB, showtime(), cpos*100/NAT_CACHE, iface, RST);

      for (k = 0; k < NAT_CACHE; k++) {

        if (nat[k].prob) {
          iaddr = (u_char *)&nat[k].ip;

          memset(host, '\0', sizeof(host));
          memset(hbar, '\0', sizeof(hbar));

          snprintf(host, sizeof(host), "%u.%u.%u.%u",
                   iaddr[0], iaddr[1], iaddr[2], iaddr[3]);

          memset(hbar, '#', nat[k].prob*2/5);

          printf(" %-15s | %u system(s) [%-40s] %3u%%\n",
                 host, nat[k].oscount, hbar, nat[k].prob);
        }
      } /* for() */

      if (cfg.verbose) {
	memset(host, '\0', sizeof(host));

	snprintf(host, sizeof(host), "%u.%u.%u.%u",
                 saddr[0], saddr[1], saddr[2], saddr[3]);

        printf("%s%sPacket SYN %-15s: TTL=%3u  TSTAMP=0x%08x  OSGENRE=%-16s%s",
               DWN, YNB, host, ip->ttl, timestamp, osgenre, RST);
      }

    } else
#endif
    if (fcount >= MIN_FACTORS    &&
        nat[i].prob >= MIN_PROB  &&
        oldprob < nat[i].prob    &&
        (nat[i].delay + REP_DELAY <= cur_time || !nat[i].delay)) {

      printf("[*] %s: NAT at %u.%u.%u.%u for %u system(s) [%u%%]\n",
             showtime(), saddr[0], saddr[1], saddr[2], saddr[3],
             nat[i].oscount, nat[i].prob);

      if (cfg.verbose) {
        if (nat[i].ttl > ip->ttl) {
          if (nat[i].tstamp > 360000)
            printf("    Gateway uptime: ~%u hour(s)\n", nat[i].tstamp/360000);
        } else {
          if (timestamp > 360000)
            printf("    Gateway uptime: ~%u hour(s)\n", timestamp/360000);
        }

        printf("    Used factors: ");
        if (factor & OSGENRE) printf("OSGENRE ");
        if (factor & TSTAMP) printf("TSTAMP ");
        if (factor & TTL) printf("TTL ");
        putchar('\n');
      }

      /* see doc/debug-mode for details */
      if (cfg.verbose == 2)
        printf("    [#%u][\"%s\", %u, %u (%lusec)] vs [\"%s\", %u, %u]\n",
               i, nat[i].osgenre, nat[i].ttl, nat[i].tstamp,
               cur_time-nat[i].time, osgenre, ip->ttl, timestamp);

      nat[i].delay = cur_time;

      goto end;
    }

    if (nat[i].ip == ip->saddr) goto end;

  } /* loop */

  if (cpos == NAT_CACHE) cpos = 0;

  /* so, we have new packet to remember */
  nat[cpos].ttl     = ip->ttl;
  nat[cpos].tstamp  = timestamp;
  nat[cpos].ip      = ip->saddr;
  nat[cpos].time    = cur_time;
  nat[cpos].delay   = 0;
  nat[cpos].oscount = 1;
  strncpy(nat[cpos].osgenre, osgenre, sizeof(nat[cpos].osgenre));
  memset(nat[cpos].oslist, 0, 2*OS_LIMIT);
  nat[cpos].oslist[0] = osint;
  cpos++;

end: ;

}


static void usage(u_char *exec) {
  fprintf(stderr,
#ifdef UNIX
  "Usage: natdet [ -Idpv ] [ -i <device> ] [ -l <file> ] [ 'filter rule' ]\n\n"
#else
  "Usage: natdet [ -Wpv ] [ -i <device> ] [ -l <file> ] [ \"filter rule\" ]\n\n"
#endif
  "  -l <file>     - dump output to this logfile\n"
  "  -i <device>   - use this interface\n"
#ifdef UNIX
  "  -I            - switch to \"I-hate-logs!\" mode\n"
  "  -d            - run in background\n"
#else
  "  -W            - list available devices\n"
#endif
  "  -p            - open interface in PROMISC mode\n"
  "  -v            - verbose output (useful)\n"
  "  -vv           - more verbose (for debugging)\n\n"
#ifdef UNIX
  "Example: %s -v -l natdet.log -i eth0 -d\n\n",
#else
  "Example: %s -v -l natdet.log -i 1\n\n",
#endif
  exec);

  exit(0);
}


#ifdef UNIX
static void int_sig() {
  fprintf(stderr, "[-] caught SIGINT(2) signal, exiting...\n");
  exit(2);
}


static void term_sig() {
  fprintf(stderr, "[-] caught SIGTERM(15) signal, exiting...\n");
  exit(15);
}


static void segv_sig() {
  fprintf(stderr, "[-] caught SIGSEGV signal, Segmentation fault!\n"
                  "    (please report it to the author)\n");
  exit(11);
}
#endif


int main(int argc, char *argv[]) {
#ifndef UNIX
  pcap_if_t *alldevs, *d;
  int adapter;
#endif
  u_char errbuff[PCAP_ERRBUF_SIZE],
         buff[256], i;

  struct bpf_program filter;
 
  pcap_t *pt;
#ifdef UNIX
  pid_t  pid;

  signal(SIGSEGV, &segv_sig);
  signal(SIGINT, &int_sig);
  signal(SIGTERM, &term_sig);
#endif

  for (i = 1; i < argc; i++) {

    if (!strcmp("-v", argv[i])) {
      cfg.verbose = 1;

    } else if (!strcmp("-vv", argv[i])) {
      cfg.verbose = 2;

    } else if (!strcmp("-p", argv[i])) {
      cfg.promisc = 1;

#ifdef UNIX
    } else if (!strcmp("-d", argv[i])) {
      cfg.detach  = 1;

    } else if (!strcmp("-I", argv[i])) {
      cfg.iactive = 1;
#else
    } else if (!strcmp("-W", argv[i])) {
      if (pcap_findalldevs(&alldevs, errbuff) == -1)
        bug("pcap_findalldevs: %s\n", errbuff);
      for(i = 1, d = alldevs; d ; d = d->next, i++) {
        printf("%d. %s ", i, d->name);
        if (d->description) printf(" (%s)\n", d->description);
        else printf(" (No description available)\n");
      }
      pcap_freealldevs(alldevs);
      exit(0);
#endif

    } else if (!strcmp("-h", argv[i]) || !strcmp("--help", argv[i])) {
      usage(argv[0]);

    } else if (!strcmp("-i", argv[i]) && i < argc - 1) {
      i++;
      iface = argv[i];

    } else if (!strcmp("-l", argv[i]) && i < argc - 1) {
      i++;
      cfg.logfile = 1;
      if (!freopen(argv[i], "a", stdout))
	bug("unable to open '%s'\n", argv[i]);
      setvbuf(stdout, (char *)NULL, _IONBF, 0);
      printf("NATDet %s started %s\n", VERSION, showtime());

    } else if (!strstr(argv[i], " ")) {
      usage(argv[0]);
    }

  }

  if (!cfg.logfile) {
    printf("NATDet - NAT Detection tool, version "VERSION"\n"
	   "(c) Marcin Ulikowski <elceef@itsec.pl>\n");
#ifndef UNIX
    printf("win32-port K. Lowczak <projectorsoft@poczta.fm>\n");
#endif
  }

#ifdef UNIX
  if (geteuid())
    bug("euid=0 is required (current=%d)\n", geteuid());
#endif

  if (strstr(argv[argc - 1], " ")  &&
      strcmp(argv[argc - 2], "-i") &&
      strcmp(argv[argc - 2], "-l")) {
    snprintf(buff, sizeof(buff), "tcp[13] == 2 and %s", argv[argc - 1]);
    rule = buff;
  } else {
    rule = "tcp[13] == 2";
  }
 

#ifdef UNIX
  if (!iface)
    iface = pcap_lookupdev(errbuff);
#else
  if (pcap_findalldevs(&alldevs, errbuff) == -1)
    bug("pcap_findalldevs: %s\n", errbuff);

  if (!iface) {
    d = alldevs;
  } else {
    adapter = atoi(iface);
    for (i = 1, d = alldevs; adapter && i < adapter && d; i++, d = d->next);
    if (!d) bug("unable to find interface %d\n", adapter);
  }

  iface = d->name;
#endif

  if (!(pt = pcap_open_live(iface, PACKET_LEN, cfg.promisc, 1, errbuff)))
    bug("unable to open device \"%s\"\n", iface);

  set_fhlen(pcap_datalink(pt));

  load_signatures();

  printf("[+] device: \"%s\", signatures: %d.\n", iface, sigs);

  if (strcmp("tcp[13] == 2", rule))
    printf("[+] filter: \"%s\"\n", rule);

  if (pcap_compile(pt, &filter, rule, 1, 0))
    bug("syntax error in filter rule.\n");

#ifdef UNIX
  if (cfg.detach && cfg.logfile && !cfg.iactive) {
    pid = fork();
    if (pid < 0) bug("cannot fork()\n");
    if (pid) exit(0);
    dup2(1, 2);
    close(0);
    setsid();
    signal(SIGHUP, SIG_IGN);
    fflush(0);

  } else if (cfg.detach && !cfg.logfile) {
    bug("switch -d requires -l <file>\n");

  } else if (cfg.detach && cfg.iactive) {
    bug("can't use both -I and -d, use screen(1) instead\n");
  }
#endif

  pcap_setfilter(pt, &filter);

#ifdef UNIX
  if (cfg.iactive) printf(CLR);
#endif

  pcap_loop(pt, -1, (pcap_handler)&watch, 0);

  pcap_close(pt);

  return 0;

}

